# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from universa.types import PrivateKey, PublicKey, KeyAddress


def test():
    private = PrivateKey(size=2048)
    private_2 = PrivateKey(key=private.key)
    assert private.equals(private_2), 'equals for private keys failed'
    assert private.key == private_2.key, 'private keys are different'

    public = private.public_key
    public_3 = PublicKey(key=private.public_key.key)
    assert public_3.equals(public), 'equals for public keys failed'
    assert public_3.key == public.key, 'public keys are different'

    key_address = public.short_address
    assert not key_address.is_long, 'short address should be `not is_long`'

    key_address_4 = KeyAddress(address=key_address.address)
    assert key_address_4.equals(key_address), 'equals for key addresses failed'
    assert key_address_4.is_matching_key_address(key_address), 'is_matching_key_address failed'
    assert key_address_4.address == key_address.address, 'key addresses are different'
    assert key_address_4.uaddress == key_address.uaddress, 'key addresses are different'

    key_address_5 = KeyAddress(uaddress=key_address.uaddress)
    assert key_address_5.equals(key_address), 'equals for key addresses failed'

    long_address = public.long_address
    assert long_address.is_long, 'long address should be is_long'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from universa.types import PrivateKey, SimpleRole, HashSet


def test():
    private_key = PrivateKey(size=2048)
    short_address = private_key.public_key.short_address
    role = SimpleRole('role', [short_address])
    x = role._invoke('getSimpleKeyAddresses')

    hashset = HashSet.get(x['id'])

    assert hashset.items[0].equals(short_address)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json
import weakref

import pexpect

try:
    from farcall import Farcall
except ImportError:
    Farcall = None

from universa import logging
from universa import exceptions

logger = logging.getLogger()


class Transport(object):
    BINARY = './umi/bin/umi'
    OBJECTS = {}

    __instance = None

    def __init__(self):
        if Transport.__instance is not None:
            logger.exception('Universa Transport is a singleton.')
            raise Exception('Universa Transport is a singleton.')

        Transport.__instance = self
        self.OBJECTS = weakref.WeakValueDictionary()
        self.__proc = None
        self.serial = 0

    def __del__(self):
        logger.info('Cleaning')
        self.drop_objects(drop_all=True)

    @staticmethod
    def get_instance():
        if Transport.__instance is None:
            Transport()
        return Transport.__instance

    @property
    def transport(self):
        if self.__proc is not None:
            return self.__proc
        self.__proc = pexpect.spawn(self.BINARY, timeout=None)
        return self.__proc

    def _format(self, **kwargs):
        return json.dumps(kwargs, separators=(',', ':'))

    def sync_call(self, name, **kwargs):
        cmd = self._format(serial=self.serial, cmd=name, **kwargs)
        logger.info('   >> executing cmd: %s', cmd)
        self.transport.sendline(cmd)
        self.transport.expect('\r\n.*"ref":%s.*\r\n' % self.serial)
        self.serial += 1
        rsp_text = self.transport.after.decode().strip()
        rsp = json.loads(rsp_text)
        logger.info('   << response: %s', rsp)
        if 'error' in rsp:
            logger.exception('Universa exception caught: %s', rsp_text)
            raise exceptions.UniversaException(rsp_text, rsp['error'])
        return rsp['result']

    def version(self):
        return self.sync_call('version')

    def instantiate(self, object_type, *args):
        return self.sync_call('instantiate', args=[object_type] + list(args))

    def invoke(self, remote_object_id, method_name, *args):
        return self.sync_call('invoke', args=[remote_object_id, method_name] + list(args))

    def invoke_static(self, object_type, method_name, *args):
        return self.sync_call('invoke', args=[object_type, method_name] + list(args))

    def get(self, remote_object_id):
        return self.sync_call('get', args=[remote_object_id])

    def drop_objects(self, ids=None, drop_all=False):
        ids = ids if ids is not None else []
        to_delete = set()
        if not drop_all:
            for _id in ids:
                if _id in self.OBJECTS:
                    to_delete.add(_id)
                    self.OBJECTS.pop(_id, None)
        else:
            to_delete = list(self.OBJECTS.keys())
            self.OBJECTS = weakref.WeakValueDictionary()

        if to_delete:
            return self.sync_call('drop_objects', ids=list(to_delete))


transport = Transport()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json

from .base import Base


class Client(Base):
    JAVA_CLASS = 'com.icodici.universa.node2.network.Client'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import warnings
from datetime import timedelta

from universa.types import PrivateKey, Contract, SimpleRole, RoleLink


def test_contracts():
    private_key = PrivateKey(size=2048)

    short_address = private_key.public_key.short_address

    sr = SimpleRole('name', [short_address])

    contract = Contract()

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        contract.set_creator_keys(private_key)
        contract.set_issuer_keys(private_key.public_key)

        assert len(w) == 4
        assert issubclass(w[0].category, PendingDeprecationWarning)  # .set_creator_keys will be deprecated
        assert issubclass(w[1].category, SyntaxWarning)  # You shouldn't use PrivateKey for roles! Never!
        assert issubclass(w[2].category, PendingDeprecationWarning)  # .set_issuer_keys will be deprecated
        assert issubclass(w[3].category, PendingDeprecationWarning)  # Use KeyAddress instead of PublicKey

    contract.set_owner_addresses(private_key.public_key.short_address)
    contract.set_owner_addresses(private_key.public_key.short_address.address)
    contract.set_owner_addresses(private_key.public_key.short_address.uaddress)

    assert contract.expires_at is None, 'expires_at should be None'
    contract.expires_at = contract.created_at + timedelta(days=90)
    assert (contract.expires_at - contract.created_at).days == 90, 'expires_at should be None'

    sealed = contract.seal()




    # contract_2 = Contract(binary=sealed)
    # assert contract_2.equals(contract), 'contracts should be equal'
    #
    #
    # simple_role = contract_2.get_owner()
    #
    # contract_3 = simple_role.get_contract()
    # assert contract_3.equals(contract), 'contracts should be equal'


def test_roles():
    simple_role = SimpleRole('my role')
    assert simple_role.get_contract() is None, 'contract should be None for role without a contract'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import warnings

from .base import Base
from .crypto import KeyAddress, PrivateKey, PublicKey
from .roles import Role
from .permissions import Permission
from universa.exceptions import UniversaException
from universa.utils import ut, dt


class Contract(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.Contract'
    ROLES = ['creator', 'issuer', 'owner']

    def __init__(self, binary=None, **kwargs):
        self._binary = binary
        super(Contract, self).__init__(**kwargs)

    @classmethod
    def from_packed(cls, from_packed):
        raise NotImplementedError()

    def _instantiate_data(self):
        if self._binary is not None:
            return {'__type': 'binary', 'base64': self._binary}

    def __set_role(self, role_name, *addresses):
        """
        Set provided addresses to the role.
        It's recommended to use KeyAddress instances, not PublicKey.
        PrivateKey is not allowed!

        :type addresses: list[KeyAddress | str]
        :returns: Role
        """
        if role_name not in self.ROLES:
            return

        _set = set()
        for address in addresses:
            __addr = None
            if isinstance(address, KeyAddress):
                __addr = address
            elif isinstance(address, PublicKey):
                warnings.warn('It\'s recommended to use KeyAddress instances, not PublicKey.',
                              PendingDeprecationWarning)
                __addr = address.short_address
            elif isinstance(address, PrivateKey):
                warnings.warn('PrivateKey instance is not allowed! Use KeyAddress instead.',
                              SyntaxWarning)
                __addr = address.public_key.short_address
            elif isinstance(address, str):
                try:
                    __addr = KeyAddress(address=address)
                except UniversaException as e:
                    if 'Not a Base58 input' in e.error.get('text', ''):
                        try:
                            __addr = KeyAddress(uaddress=address)
                        except UniversaException:
                            pass

            if __addr is not None:
                _set.add(__addr)

        cmd = 'set{}Keys'.format(role_name.capitalize())
        rsp = self._invoke(cmd, *[__addr.remote for __addr in _set])
        klass = self.get_class_by_java(rsp['className'])
        return klass.get(rsp['id'])

    def set_creator_keys(self, *keys):
        """Set 'creator' role to given keys."""
        warnings.warn('Contract.set_creator_keys(*keys) will be deprecated. '
                      'Use contract.set_creator_addresses(*addresses) instead.',
                      PendingDeprecationWarning)
        return self.set_creator_addresses(*keys)

    def set_issuer_keys(self, *keys):
        """Set 'issuer' role to given keys."""
        warnings.warn('Contract.set_issuer_keys(*keys) will be deprecated. '
                      'Use contract.set_issuer_addresses(*addresses) instead.',
                      PendingDeprecationWarning)
        return self.set_issuer_addresses(*keys)

    def set_owner_keys(self, *keys):
        """Set 'owner' role to given keys."""
        warnings.warn('Contract.set_owner_keys(*keys) will be deprecated. '
                      'Use contract.set_owner_addresses(*addresses) instead.',
                      PendingDeprecationWarning)
        return self.set_owner_addresses(*keys)

    def set_creator_addresses(self, *addresses):
        """Set 'creator' role to given addresses."""
        return self.__set_role('creator', *addresses)

    def set_issuer_addresses(self, *addresses):
        """Set 'issuer' role to given addresses."""
        return self.__set_role('issuer', *addresses)

    def set_owner_addresses(self, *addresses):
        """Set 'owner' role to given addresses."""
        return self.__set_role('owner', *addresses)

    def get_role(self, role_name):
        if role_name in self.ROLES:
            rsp = self._invoke('get{}'.format(role_name.capitalize()))
        else:
            rsp = self._invoke('getRole', role_name)

        if rsp is not None:
            klass = self.get_class_by_java(rsp['className'])
            return klass.get(rsp['id'])
        else:
            return None

    def get_creator(self):
        return self.get_role('creator')

    def get_issuer(self):
        return self.get_role('issuer')

    def get_owner(self):
        return self.get_role('owner')

    def create_role(self, name, role):
        """
        :type name: str
        :type role: Role
        :return: Role
        """
        rsp = self._invoke('createRole', name, role.remote)
        klass = self.get_class_by_java(rsp['className'])
        return klass.get(rsp['id'])

    def get_created_at(self):
        """
        :return: unix time
        :rtype: datetime | None
        """
        rsp = self._invoke('getCreatedAt')
        return dt(rsp['seconds'])
    created_at = property(get_created_at)

    def get_expires_at(self):
        """
        :return: unix time
        :rtype: datetime | None
        """
        rsp = self._invoke('getExpiresAt')
        return dt(rsp['seconds']) if rsp is not None else None

    def set_expires_at(self, expires_at):
        """
        :param expires_at: unix time
        :type expires_at: datetime
        """
        self._invoke('setExpiresAt', {'seconds': ut(expires_at), '__type': 'unixtime'} 
                                         if expires_at is not None 
                                         else None)
    expires_at = property(get_expires_at, set_expires_at)

    def add_permission(self, permission):
        """
        :type permission: Permission
        """
        self._invoke('addPermission', permission.remote)

    def get_permissions(self):
        raise NotImplementedError()
        rsp = self._invoke('getPermissions')

    def is_permitted(self, permission_name, *keys):
        """
        Checks if permission of given type that is allowed for given keys exists.

        :type permission_name: str
        :type keys: PublicKey
        :rtype: bool
        """
        return self._invoke('isPermitted', permission_name, [key.remote for key in keys])

    def check_applicable_permission_quantized(self, permission):
        """
        Quantize given permission (add cost for that permission).

        :type permission: Permission
        """
        self._invoke('checkApplicablePermissionQuantized', permission.remote)

    def seal(self):
        return self._invoke('seal')['base64']

    def check(self):
        return self._invoke('check')

    def set_state(self):
        raise NotImplementedError()

    def get_processed_cost(self):
        return self._invoke('getProcessedCost')

    def get_processed_cost_u(self):
        return self._invoke('getProcessedCostU')


class ContractsService(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.ContractsService'


class Reference(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.Reference'


class Parcel(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.Parcel'


class ExtendedSignature(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.ExtendedSignature'




# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json

from .base import transport, Base


class HashSet(Base):
    JAVA_CLASS = 'java.util.HashSet'

    def __init__(self, items=None, **kwargs):
        __items = set()
        for item in items or []:
            obj = None
            klass = self.get_class_by_type(item['__type'])
            if klass.__name__ == 'KeyAddress':
                obj = klass(uaddress=item['uaddress']['base64'])

            if obj is not None:
                __items.add(obj)

        super(HashSet, self).__init__(**kwargs)
        self.items = list(__items)

    @classmethod
    def get(cls, _id):
        instance = cls(id=_id, items=cls._get(_id), skip_update=True)
        transport.OBJECTS[_id] = instance
        return instance

    def update(self):
        self.items = self._get(self.id)

    def _instantiate_data(self):
        return



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import gc

from universa.transport import transport
from universa.types import PrivateKey


def test():
    gc.enable()

    assert len(transport.OBJECTS) == 0, 'transport objects are not blank'

    key = PrivateKey(size=2048)
    assert len(transport.OBJECTS) == 1, 'transport objects does not contains only one value'

    key2 = PrivateKey(size=2048)
    assert len(transport.OBJECTS) == 2, 'transport objects does not contains only two values'

    # Create a key overriding previous one variable.
    # GarbageCollector should delete old object and kill it from transport (and UMI remote object).
    key2 = PrivateKey(size=2048)
    assert len(transport.OBJECTS) == 2, 'transport objects does not contains only two values'

    # Create a key and don't save it to the variable. It should be deleted by UMI at once.
    PrivateKey(size=2048)
    assert len(transport.OBJECTS) == 2, 'transport objects does not contains only two values'

    # Delete a local key object. It should be removed from UMI too.
    del key2
    assert len(transport.OBJECTS) == 1, 'transport objects does not contains only one value'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json

from .base import Base


class HashId(Base):
    JAVA_CLASS = 'com.icodici.universa.HashId'

    def __init__(self, packed_data=None, **kwargs):
        # TODO: check this constructor
        raise NotImplementedError()
        self.packed_data = packed_data
        super(HashId, self).__init__(**kwargs)

    def _instantiate_data(self):
        return self.packed_data

    @staticmethod
    def from_get(data):
        return {'packed_data': data['composite3']}

    def compare_to(self, obj):
        return self._equals('compareTo', obj)

    @classmethod
    def create_random(cls):
        return cls.get(cls._invoke_static('createRandom')['id'])

    def get_digest(self):
        return self._invoke('getDigest')

    @property
    def hash_code(self):
        return self._invoke('hashCode')

    def to_base64_string(self):
        return self._invoke('toBase64String')

    @classmethod
    def with_digest(cls, _hash=None, encoded=None):
        assert (_hash is None) ^ (encoded is None), 'either hash or encoded should be not None'
        rsp = cls._invoke_static('withDigest', _hash or encoded)
        return cls.get(rsp['id'])



import setuptools


CLASSIFIERS = [
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Developers',
]


with open('README.md', 'r') as fh:
    long_description = fh.read()


setuptools.setup(
    name='universa',
    version='0.0.2',
    author='Vadim Kovrigin',
    author_email='kovrigin.dev@gmail.com',
    description='Python package to access Universa API from python.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['Universa'],
    url='https://github.com/vkovrigin/universa',
    packages=setuptools.find_packages(),
    classifiers=CLASSIFIERS,
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import calendar
from datetime import datetime


def ut(date_time):
    return calendar.timegm(date_time.timetuple())


def dt(unix_time):
    return datetime.utcfromtimestamp(float(unix_time))

